import os
import shutil
from tal.config import local_file_path
from tal.util import fdent
from textwrap import indent
import tal
import datetime

from tal.render.util import import_mitsuba_backend
mitsuba_backend = import_mitsuba_backend()


def create_nlos_scene(folder_name, args):
    folder_name = os.path.abspath(folder_name)
    path, name = os.path.split(folder_name)

    assert os.path.isdir(path), f'{path} does not exist'
    if os.path.exists(folder_name):
        answer = None
        num_files = sum(map(lambda e: len(e[2]), os.walk(folder_name)))
        while answer is None:
            answer = input(
                f'WARNING: {folder_name} already exists and contains {num_files} file{"" if num_files == 1 else "s"}.\n'
                f'Erase ALL and replace? (y/N): ')
        print()
        if answer.lower() != 'y' and answer.lower() != 'yes':
            print('Operation cancelled.')
            exit()
        else:
            shutil.rmtree(folder_name)

    try:
        os.mkdir(folder_name)
    except OSError as exc:
        raise AssertionError(f'Invalid permissions: {exc}') from exc

    config_name = os.path.join(folder_name, f'{name}.yaml')
    default_yaml_data = open(local_file_path(
        'render/scene_defaults.yaml'), 'r').readlines()
    first_uncomment = default_yaml_data.index('geometry:\n')
    data_commented, data_uncommented = \
        default_yaml_data[:first_uncomment], default_yaml_data[first_uncomment:]
    data_commented = ''.join(list(
        map(lambda l: l.replace(r'$default_mitsuba_variant$', mitsuba_backend.get_default_variant()),
            map(lambda l: l if len(l.strip()) == 0 or l.startswith('#') else f'#{l}', data_commented))))
    default_yaml_data = ''.join([*data_commented, *data_uncommented])

    def print_material(item):
        key, value = item

        material_keys = list(
            map(lambda s: f'{s}: ${s}', mitsuba_backend.get_material_keys(value)))
        value = indent(value, '#| ')

        if len(material_keys) == 0:
            return fdent('''\
                #| id: {key}
                {value}''', key=key, value=value)
        else:
            material_keys = indent('\n'.join(material_keys), '#| ')
            return fdent('''\
                #| id: {key}
                {material_keys}
                {value}''', key=key, material_keys=material_keys, value=value)

    default_material_data = '\n\n'.join(
        map(print_material, mitsuba_backend.get_materials().items()))

    with open(config_name, 'w') as f:
        f.write(
            '# TAL v{v} NLOS scene description file: {r}\n'.format(
                v=tal.__version__, r='https://github.com/diegoroyo/tal'))
        f.write('# Created on {d} with experiment name "{n}"\n'.format(
                d=datetime.datetime.now().strftime(r'%Y/%m/%d %H:%M'), n=name))
        f.write('#\n')
        f.write('# All commented variables show their default value\n')
        f.write('# To render the scene with this configuration, execute:\n')
        f.write('#   [python3 -m] tal render {c}\n'.format(
                c=config_name))
        f.write('name: {name}\n'.format(name=name))
        f.write('\n')
        f.write(default_yaml_data)
        f.write(default_material_data)

    if not args.quiet:
        print(f'Success! Now:\n'
              f'1) Edit the configuration file in {config_name}\n'
              f'2) Render with tal render {config_name}')



import dill
import os
import shutil
import yaml
import multiprocessing as mp
import numpy as np
from tqdm import tqdm
from enum import Enum

# FIXME decide if we keep this here or not
import resource
import psutil
import gc

from tal.util import local_file_path

""" Utils to read/write to persistent config file """


class Config(Enum):
    MITSUBA_VERSION = \
        ('MITSUBA_VERSION',
         'Version of Mitsuba to use (2 or 3)',
         '2',
         lambda s: s in ['2', '3'])
    MITSUBA2_TRANSIENT_NLOS_FOLDER = \
        ('MITSUBA2_TRANSIENT_NLOS_FOLDER',
         'Location of mitsuba2-transient-nlos installation folder',
         '',
         lambda s: os.path.isdir(s))
    MITSUBA3_TRANSIENT_NLOS_FOLDER = \
        ('MITSUBA3_TRANSIENT_NLOS_FOLDER',
         'Location of mitsuba3-transient-nlos installation folder',
         '',
         lambda s: os.path.isdir(s))


def get_home_path():
    return os.path.expanduser('~')


def get_config_filename():
    return os.path.expanduser('~/.tal.conf')


def _parse_config(lines):
    return dict(line.strip().split('=') for line in lines)


def ask_for_config(param_name: Config, force_ask=True):
    """Most useful: ask for a specific key in the config,
    and if it does not exist ask the user for a value"""
    config_dict = read_config()
    param_name, ask_query, default_value, is_valid = param_name.value
    param_ok = (
        param_name in config_dict and
        len(config_dict[param_name].strip()) > 0 and
        is_valid(config_dict[param_name])
    )
    default_value = (
        config_dict[param_name]
        if param_ok and len(default_value) == 0
        else default_value
    )
    if force_ask or not param_ok:
        print(f'{param_name} is not specified. Please specify:')
        param_value = ''
        while len(param_value.strip()) == 0 or not is_valid(param_value):
            param_value = input(
                f'{ask_query}{"" if len(default_value) == 0 else f" [default: {default_value}]"}: ')
            param_value = param_value.strip()
            if len(param_value) == 0:
                param_value = default_value
            if not is_valid(param_value):
                print('Invalid value.')
        config_dict[param_name] = param_value
        write_config(config_dict)
    return config_dict[param_name]


def read_config() -> dict:
    config = get_config_filename()
    if not os.path.isfile(config):
        print('TAL configuration file not found. '
              f'Creating a new one in {config}...')
        try:
            shutil.copy(local_file_path('.tal.conf.example'), config)
        except Exception as exc:
            print(f'/!\\ Unknown error when creating config file: {exc}')

    with open(config, 'r') as f:
        return _parse_config(f.readlines())


def write_config(config_dict):
    config = get_config_filename()
    with open(config, 'w') as f:
        f.writelines(list(f'{key}={value}\n'
                          for key, value in config_dict.items()))


def read_yaml(filename: str):
    with open(filename, 'r') as f:
        try:
            return yaml.safe_load(f)
        except yaml.YAMLError as exc:
            raise AssertionError(
                f'An error ocurred when parsing file ({filename}):\n{exc}')


def write_yaml_string(data_dict: dict) -> str:
    return yaml.dump(data_dict)


""" Utils to read/write to the context configuration (tal.config) """


def _run_dill_encoded(payload):
    fun, args = dill.loads(payload)
    return fun(*args)


def _apply_async(pool, fun, args, **kwargs):
    payload = dill.dumps((fun, args))
    return pool.apply_async(_run_dill_encoded, (payload,), **kwargs)


DEFAULT_CPU_PROCESSES = 1
DEFAULT_DOWNSCALE = None
DEFAULT_CALLBACK = None


class ResourcesConfig:
    """ See tal.set_resources """

    def __init__(self,
                 cpu_processes=DEFAULT_CPU_PROCESSES,
                 downscale=DEFAULT_DOWNSCALE,
                 callback=DEFAULT_CALLBACK):
        self.cpu_processes = cpu_processes
        self.downscale = downscale
        self.callback = callback
        self.force_single_thread = cpu_processes == 1 and downscale is None

    def __enter__(self):
        global TAL_RESOURCES_CONFIG
        self.old_resources = TAL_RESOURCES_CONFIG
        TAL_RESOURCES_CONFIG = self

    def __exit__(self, _, __, ___):
        global TAL_RESOURCES_CONFIG
        TAL_RESOURCES_CONFIG = self.old_resources
        del self.old_resources
        return False

    def split_work(self, f_work, data_in, data_out, slice_dims):

        # FIXME decide what to do with memory management
        # and also with other parameters' names
        # gc.collect()
        # _, hard = resource.getrlimit(resource.RLIMIT_AS)
        # free_memory_bytes = int(psutil.virtual_memory().free * 0.98)
        # free_memory_gb = free_memory_bytes / 2 ** 30
        # print('tal.resources: Setting memory limit to '
        #       f'{free_memory_gb:.2f} GiB.')
        # resource.setrlimit(resource.RLIMIT_AS, (free_memory_bytes, hard))

        def single_process():
            data_out[:] = f_work(data_in)

        if self.force_single_thread:
            single_process()
            return

        in_slice_dim, out_slice_dim = slice_dims

        def check_divisible(x):
            if in_slice_dim is not None and data_in.shape[in_slice_dim] % x != 0:
                return False
            if out_slice_dim is not None and data_out.shape[out_slice_dim] % x != 0:
                return False
            return True

        cpus = min(
            mp.cpu_count(),
            999 if self.cpu_processes == 'max' else self.cpu_processes
        )
        max_downscale = 128
        if self.downscale is None:
            downscale = 1
        else:
            downscale = 2**int(np.ceil(np.log2(self.downscale)))
        downscale = min(max_downscale, downscale)
        while downscale < cpus and downscale < max_downscale and check_divisible(downscale * 2):
            downscale *= 2

        data_dim = min((x for x in [
                        None if in_slice_dim is None else data_in.shape[in_slice_dim],
                        None if out_slice_dim is None else data_out.shape[out_slice_dim],
                        ] if x is not None), default=None)

        if downscale == 1:
            single_process()
            return

        print(f'tal.resources: Using {cpus} CPU processes '
              f'and downscale {downscale}.')

        if in_slice_dim is not None:
            in_shape = np.insert(data_in.shape, in_slice_dim + 1, downscale)
            in_shape[in_slice_dim] //= downscale
        if out_slice_dim is not None:
            out_shape = np.insert(data_out.shape, out_slice_dim + 1, downscale)
            out_shape[out_slice_dim] //= downscale

        with mp.Pool(processes=cpus) as pool:
            try:
                def do_work(din):
                    return _apply_async(pool, f_work, (din,),
                                        error_callback=lambda exc: print(f'/!\ Process found an exception: {exc}'))

                if in_slice_dim is not None:
                    data_in = np.moveaxis(
                        data_in.reshape(in_shape), in_slice_dim + 1, 0)
                if out_slice_dim is not None:
                    data_out = np.moveaxis(
                        data_out.reshape(out_shape), out_slice_dim + 1, 0)

                asyncs = []
                for in_slice in data_in:
                    asyncs.append(do_work(in_slice))

                pool.close()
                tqdm_kwargs = {
                    'total': len(asyncs),
                    'desc': 'tal.resources progress',
                }
                if out_slice_dim is None:
                    for i, async_ in tqdm(enumerate(asyncs), **tqdm_kwargs):
                        data_out += async_.get()
                        if self.callback is not None:
                            self.callback(data_out, i, len(asyncs))
                else:
                    for i, (async_, out_slice) in tqdm(enumerate(zip(asyncs, data_out)), **tqdm_kwargs):
                        out_slice[:] = async_.get()
                        if self.callback is not None:
                            self.callback(data_out, i, len(asyncs))
                pool.join()
            # FIXME decide if this should be here
            # except MemoryError:
            #     pool.terminate()
            #     pool.join()
            #     raise MemoryError(f'tal.resources: Memory error, {free_memory_gb:.2f} GiB is not enough.'
            #                       ' Either decrease CPU processes, increase downscale, or move to another system.')
            except KeyboardInterrupt:
                pool.terminate()
                pool.join()
                raise KeyboardInterrupt


TAL_RESOURCES_CONFIG = ResourcesConfig()


def get_resources():
    return TAL_RESOURCES_CONFIG


def set_resources(cpu_processes=DEFAULT_CPU_PROCESSES, downscale=DEFAULT_DOWNSCALE, callback=DEFAULT_CALLBACK):
    """
    Configures how Y-TAL should execute your code so you can manage CPU/RAM use.

    Not all functions implemented in Y-TAL support this configuration, mostly the ones
    in tal.reconstruct (filtering and reconstruction).

    Default configuration is to process all the data simultaneously in 1 CPU process.

    cpu_processes
        Can be an integer or 'max' to use all available CPU processes.

    downscale
        Can be an integer or None.
        If integer, the data will be processed in smaller chunks (e.g. if downscale = 8, the
            data will be processed 1/8th at a time).
        If None, all the data will be processed at the same time.
        In any case, downscale will be increased automatically make sure that all cpu_processes
            can run at the same time.

    callback
        A function that will be called after each chunk of data is processed.
        It will receive (1) the data, (2) the current chunk index and (3) the total number of chunks as arguments.
        This is useful to show a progress bar, or partial reconstructions for example.
        See tal.callbacks for some default implementations which may or may not work for your case.
        e.g. callback=lambda x: `tal.plot.amplitude_phase(x[0], title=f'Chunk {x[1]} of {x[2]}')`
    """
    global TAL_RESOURCES_CONFIG
    TAL_RESOURCES_CONFIG = ResourcesConfig(cpu_processes, downscale, callback)



"""
tal.plot
========

Functions for plotting and examining captured data or reconstructions using that data.

They can be called from Python, but also can be called using the command line interface
by running tal plot <function_name> <args> <kwargs> (see tal plot -h).
"""

import numpy as np  # FIXME type hints for function below
from tal.io.capture_data import NLOSCaptureData
from typing import Union, List


_Data = Union[NLOSCaptureData, NLOSCaptureData.HType]
_DataList = Union[List[_Data], _Data]


def amplitude_phase(data: np.ndarray, title: str = ''):
    """
    For a 2D array with complex data, plot amplitude/phase
    """
    from tal.plot.xy import plot_amplitude_phase
    return plot_amplitude_phase(data, title)


def xy_grid(data: _Data,
            size_x: int = 8, size_y: int = 8,
            t_start: int = None, t_end: int = None, t_step: int = 1):
    """
    For a single impulse response H with shape (T, Sx, Sy), plot a grid of temporal slices.

    size_x, size_y
        Size of the grid.

    t_start, t_end, t_step
        Limit the temporal range to [t_start, t_end) with step t_step.
        This operation is done before the grid is created.
    """
    from tal.plot.xy import plot_xy_grid
    return plot_xy_grid(data, size_x, size_y, t_start, t_end, t_step)


def t_comparison(data_list: _DataList,
                 x: int = None, y: int = None,
                 t_start: int = None, t_end: int = None,
                 a_min: float = None, a_max: float = None,
                 normalize: bool = False,
                 labels: List[str] = None):
    """
    For impulse responses H, do an interactive plot of transient pixels.
    This function can be called with multiple H (to compare transient pixels),
    but they must have the same shape.

    Preferred to call using CLI e.g. tal plot t_comparison <file_1> <file_2>
    """
    from tal.plot.compare import plot_t_comparison
    return plot_t_comparison(data_list, x, y, t_start, t_end, a_min, a_max, normalize, labels)


def volume(data: _Data, title: str = '', slider_title: str = 'Time',
           slider_step: float = 0.1, color: str = 'hot',
           opacity: str = 'sigmoid', backgroundcolor: str = None):
    """ 3D volume plot of the data """
    from tal.plot.plotter3d import plot3d
    return plot3d(data, title, slider_title, slider_step, color, opacity,
                  backgroundcolor)


def xy_interactive(data: _Data, cmap: str = 'hot'):
    """ See txy_interactive. Shortcut with slice_axis='t' """
    from tal.plot.xy import plot_txy_interactive
    return plot_txy_interactive(data, cmap, 't')


def tx_interactive(data: _Data, cmap: str = 'hot'):
    """ See txy_interactive. Shortcut with slice_axis='y' """
    from tal.plot.xy import plot_txy_interactive
    return plot_txy_interactive(data, cmap, 'y')


def ty_interactive(data: _Data, cmap: str = 'hot'):
    """ See txy_interactive. Shortcut with slice_axis='x' """
    from tal.plot.xy import plot_txy_interactive
    return plot_txy_interactive(data, cmap, 'x')


def txy_interactive(data: _Data, cmap: str = 'hot', slice_axis: str = 't'):
    """
    For 3D data, plot a 2D slice of the data with the given slice_axis fixed.
    User can interactively change the slice position.
    """
    from tal.plot.xy import plot_txy_interactive
    return plot_txy_interactive(data, cmap, slice_axis)



from tal.enums import FileFormat, GridFormat, HFormat, VolumeFormat
import numpy as np


def __convert_dict_znlos_to_tal(capture_data: dict) -> dict:
    H = capture_data['data']
    if H.ndim == 7:
        # Exhaustive/single - convert to (colors, bounces, t, lx, ly, sx, sy)
        if H.shape[4] == 6:
            # assume (ly, lx, sy, sx, bounces, t, colors)
            H = np.moveaxis(H, list(range(7)), (4, 3, 6, 5, 1, 2, 0))
        elif H.shape[2] == 6:
            # assume (colors, t, bounces, sx, sy, lx, ly)
            H = np.moveaxis(H, list(range(7)), (0, 2, 1, 6, 5, 4, 3))
        else:
            raise AssertionError(
                'Conversion not detected for this H.ndim = 7 case')
    elif H.ndim == 5:
        # Confocal - convert to (colors, bounces, t, sx, sy)
        if H.shape[0] == H.shape[1]:
            # assume (sy, sx, bounces, t, colors)
            H = np.moveaxis(H, list(range(5)), (3, 4, 1, 2, 0))
        elif H.shape[3] == H.shape[4]:
            # assume (colors, t, bounces, sx, sy)
            H = np.moveaxis(H, list(range(5)), (0, 2, 1, 4, 3))
        else:
            raise AssertionError(
                'Conversion not detected for this H.ndim = 5 case')
    else:
        raise AssertionError('Conversion not implemented for H.ndim != 5 or 7')
    # sum colors and bounces dims (t, lx, ly, sx, sy)
    H = np.sum(H, axis=(0, 1))
    # remove (1, 1) dims (e.g. laser in single capture 1x1x256x256)
    H = np.squeeze(H)
    H_format = HFormat.T_Sx_Sy if H.ndim == 3 else HFormat.T_Lx_Ly_Sx_Sy

    def conv_to_xy3(arr):
        if arr.shape[0] == 3:
            return np.transpose(arr)
        else:
            return arr

    def parse_volume_size(volume_size):
        volume_size = np.array(volume_size, dtype=np.float32)
        if volume_size.size == 1:
            volume_size = np.repeat(volume_size, 3)
        return volume_size

    return {
        'H': H,
        'H_format': H_format,
        'sensor_xyz': capture_data['cameraPosition'].reshape(3),
        'sensor_grid_xyz': conv_to_xy3(capture_data['cameraGridPositions']),
        'sensor_grid_normals': conv_to_xy3(capture_data['cameraGridNormals']),
        'sensor_grid_format': GridFormat.X_Y_3,
        'laser_xyz': capture_data['laserPosition'].reshape(3),
        'laser_grid_xyz': conv_to_xy3(capture_data['laserGridPositions']),
        'laser_grid_normals': conv_to_xy3(capture_data['laserGridNormals']),
        'laser_grid_format': GridFormat.X_Y_3,
        'volume_format': VolumeFormat.X_Y_Z_3,
        'delta_t': capture_data['deltaT'],
        't_start': capture_data['t0'],
        't_accounts_first_and_last_bounces': True,
        'scene_info': {
            'original_format': 'HDF5_ZNLOS',
            'volume': {
                'center': capture_data['hiddenVolumePosition'].reshape(3),
                'rotation': capture_data['hiddenVolumeRotation'].reshape(3),
                'size': parse_volume_size(capture_data['hiddenVolumeSize']),
            }
        },
    }


def __convert_dict_dirac_to_tal(capture_data: dict) -> dict:
    t_accounts_first_and_last_bounces = np.isclose(
        0.0, np.linalg.norm(capture_data['offset_laser']))

    nx = len(capture_data['xa'])
    ny = len(capture_data['ya'])
    da = capture_data['da']
    nt = len(da)

    sensor_grid = np.stack(
        (capture_data['xg'],
         capture_data['yg'],
         np.zeros((nx, ny), dtype=np.float32)),
        axis=-1)
    laser_grid = np.copy(sensor_grid)

    def expand(vec, x, y):
        assert len(vec) == 3
        return vec.reshape(1, 1, 3).repeat(x, axis=0).repeat(y, axis=1)

    print('Warning: converting from HDF5_NLOS_DIRAC does some assumptions '
          'on the positions of the laser and sensor. If your data has '
          't_accounts_first_and_last_bounces = True, it has a chance to be wrong.')

    return {
        'H': capture_data['data_t'],
        'H_format': HFormat.T_Sx_Sy,
        'sensor_xyz': np.array([-0.5, 0, 0.25], dtype=np.float32),
        'sensor_grid_xyz': sensor_grid,
        'sensor_grid_normals': expand(np.array([0.0, 0.0, 1.0], dtype=np.float32), nx, ny),
        'sensor_grid_format': GridFormat.X_Y_3,
        'laser_xyz': np.array([-0.5, 0, 0.25], dtype=np.float32),
        'laser_grid_xyz': laser_grid,
        'laser_grid_normals': expand(np.array([0.0, 0.0, 1.0], dtype=np.float32), nx, ny),
        'laser_grid_format': GridFormat.X_Y_3,
        'volume_format': VolumeFormat.X_Y_Z_3,
        'delta_t': da[1] - da[0],
        't_start': None if t_accounts_first_and_last_bounces else 0.0,
        't_accounts_first_and_last_bounces': t_accounts_first_and_last_bounces,
        'scene_info': {
            'original_format': 'HDF5_NLOS_DIRAC',
            'offset': {
                'sensor': capture_data['offset_camera'].reshape(nx, ny),
                'laser': capture_data['offset_laser'].reshape(1, 1),
            },
            'sample_wavelength': {
                'x': capture_data['sample_x'],
                'y': capture_data['sample_y'],
            }
        },
    }


def __convert_dict_pfmat_to_tal(capture_data: dict) -> dict:
    dataset = capture_data['dataset']

    nt_expected = dataset['t']
    H = dataset['data']
    nt, ns, nl = H.shape
    assert H.ndim == 3 and nt == nt_expected and ns == 1, \
        'Conversion assumes H to be of shape (nt, ns, nl)'
    # convert to (nt, nl)
    H = np.squeeze(H)

    def conv_to_n3(arr, expected_len=None):
        if arr.shape[0] == 3:
            arr = np.transpose(arr)
        if expected_len is not None and arr.shape[0] == 1:
            arr = np.stack((arr,) * expected_len, axis=0)
        return arr

    # NOTE(diego): this data typically uses one camera position,
    # and multiple laser positions, taking advantage of the fact that
    # light propagation in both path directions (laser->sensor and sensor->laser)
    # is equivalent. We invert laser and sensors to be more consistent with
    # other dataset types and existing code
    return {
        'H': H,
        'H_format': HFormat.T_Si,
        'sensor_xyz': dataset['laserOrigin'].reshape(3),
        'sensor_grid_xyz': conv_to_n3(dataset['laserPos']),
        'sensor_grid_normals': conv_to_n3(dataset['laserNorm'], expected_len=nl),
        'sensor_grid_format': GridFormat.N_3,
        'laser_xyz': dataset['cameraPos'].reshape(3),
        'laser_grid_xyz': conv_to_n3(dataset['cameraPos']),
        'laser_grid_normals': conv_to_n3(dataset['cameraNormal'], expected_len=ns),
        'laser_grid_format': GridFormat.N_3,
        'volume_format': VolumeFormat.X_Y_Z_3,
        'delta_t': dataset['deltat'],
        't_start': dataset['t0'],
        't_accounts_first_and_last_bounces': True,
        'scene_info': {
            'original_format': 'MAT_PHASOR_FIELDS',
            'volume': {
                'minimal_pos': capture_data['minimalpos'].reshape(3),
                'maximal_pos': capture_data['maximalpos'].reshape(3),
                'sampling_grid_spacing': capture_data['sampling_grid_spacing'],
            }
        },
    }


def __convert_dict_tal_to_znlos(capture_data: dict) -> dict:
    raise NotImplementedError('Conversion to HDF5_ZNLOS not implemented')


def __convert_dict_tal_to_dirac(capture_data: dict) -> dict:
    raise NotImplementedError('Conversion to HDF5_NLOS_DIRAC not implemented')


def __convert_dict_tal_to_pfmat(capture_data: dict) -> dict:
    raise NotImplementedError(
        'Conversion to MAT_PHASOR_FIELDS not implemented')


def detect_dict_format(raw_data: dict) -> FileFormat:
    if 'data' in raw_data:
        return FileFormat.HDF5_ZNLOS
    elif 'data_t' in raw_data:
        return FileFormat.HDF5_NLOS_DIRAC
    elif 'H' in raw_data:
        return FileFormat.HDF5_TAL
    elif 'dataset' in raw_data:
        return FileFormat.MAT_PHASOR_FIELDS
    else:
        raise AssertionError('Unable to detect capture data file format')


def convert_dict(capture_data: dict,
                 format_to: FileFormat) -> dict:
    """
    Convert raw data from one format to another
    """
    # convert to HDF5_TAL
    file_format = detect_dict_format(capture_data)
    if file_format == FileFormat.HDF5_TAL:
        capture_data_tal = capture_data
    elif file_format == FileFormat.HDF5_ZNLOS:
        capture_data_tal = __convert_dict_znlos_to_tal(capture_data)
    elif file_format == FileFormat.HDF5_NLOS_DIRAC:
        capture_data_tal = __convert_dict_dirac_to_tal(capture_data)
    elif file_format == FileFormat.MAT_PHASOR_FIELDS:
        capture_data_tal = __convert_dict_pfmat_to_tal(capture_data)
    else:
        raise AssertionError(
            'convert_dict not implemented for this file format')

    # convert from HDF5_TAL to output format
    if format_to == FileFormat.HDF5_TAL:
        return capture_data_tal
    elif format_to == FileFormat.HDF5_ZNLOS:
        return __convert_dict_tal_to_znlos(capture_data_tal)
    elif format_to == FileFormat.HDF5_NLOS_DIRAC:
        return __convert_dict_tal_to_dirac(capture_data_tal)
    elif format_to == FileFormat.MAT_PHASOR_FIELDS:
        return __convert_dict_tal_to_pfmat(capture_data_tal)
    else:
        raise AssertionError(
            'convert_dict not implemented for this file format')



import argparse
import os
from tal import __version__ as tal_version
from tal.util import fdent


class SmartFormatter(argparse.HelpFormatter):
    def _split_lines(self, text, width):
        if text.startswith('R|'):
            return text[2:].splitlines()
        # this is the RawTextHelpFormatter._split_lines
        return argparse.HelpFormatter._split_lines(self, text, width)


def get_plot_functions():
    """
    Returns (function_names, function_params) where
    function_names is a list of str
    function_params is a list of tuples
        (param_name, param_type)
    """
    from inspect import getmembers, isfunction, signature
    import tal.plot
    function_names = list()
    function_param_names = dict()
    function_param_data = set()
    for name, func in getmembers(tal.plot, isfunction):
        function_names.append(name)
        parameters = signature(func).parameters
        parameter_names = list(filter(lambda p: p not in ['data', 'data_list', 'args', 'kwargs'],
                                      list(parameters)))
        function_param_names[name] = parameter_names
        function_param_data.update(
            list(map(lambda p: (p, parameters[p].annotation),
                     parameter_names)))
    return function_names, function_param_names, function_param_data


def main():
    parser = argparse.ArgumentParser(
        description=f'Y-TAL - (Your) Transient Auxiliary Library - v{tal_version}', formatter_class=SmartFormatter)
    parser.add_argument('-v', '--version', action='version',
                        version=f'%(prog)s v{tal_version}')
    subparsers = parser.add_subparsers(
        help='Command', required=True, dest='command')

    # config commands
    config_parser = subparsers.add_parser(
        'config', help='Edit the TAL configuration file')

    # render commands
    render_parser = subparsers.add_parser(
        'render', help='Create, edit or execute renders of simulated NLOS scene data captures', formatter_class=SmartFormatter)
    render_parser.add_argument('config_file', nargs='*', help=fdent('''\
                                    R|Can be:
                                    1) Path to a TAL scene config YAML file
                                    2) Path to a TAL scene directory (must have a YAML file inside with the same name as the directory)
                                    3) 'new <folder_name>' to create a new folder (i.e. tal render new <folder_name>)'''))
    render_parser.add_argument('-t', '--threads',
                               type=int, default=os.cpu_count() - 1, required=False,
                               help='Number of threads')
    render_parser.add_argument('-s', '--seed',
                               type=int, default=0, required=False,
                               help='Random seed for the sampler. Without setting this value to different values, the same results will be produced everytime.')
    render_parser.add_argument('-n', '--nice',
                               type=int, default=0, required=False,
                               help='Change +/- in nice factor. Positive values = lower priority. Negative values = higher priority (needs sudo)')
    render_parser.add_argument('-q', '--quiet',
                               dest='quiet', action='store_true',
                               help='Disable progress bars and other verbose outputs')
    render_parser.add_argument('-g', '--gpu',
                               dest='gpus', nargs='*', type=int, default=[], required=False,
                               help='Select which GPUs should be used by Mitsuba')
    render_parser.add_argument('-r', '--dry-run',
                               dest='dry_run', action='store_true',
                               help='Do not execute mitsuba, just print out the commands that would be executed')
    render_parser.add_argument('--no-steady',
                               dest='do_steady_renders', action='store_false',
                               help='Disable generation of steady state images')
    render_parser.add_argument('--no-logging',
                               dest='do_logging', action='store_false',
                               help='Disable logging of mitsuba2 output')
    render_parser.add_argument('--no-partial-results',
                               dest='keep_partial_results', action='store_false',
                               help='Remove the "partial" folder which stores temporal data after creating the final hdf5 file (e.g. multiple experiments for confocal/exhaustive)')

    # plot commands
    plot_parser = subparsers.add_parser(
        'plot', help='Plot capture data using one of the configured methods', formatter_class=SmartFormatter)
    plot_func_names, plot_func_param_names, plot_func_param_data = get_plot_functions()
    plot_parser.add_argument('preset', help=fdent('''\
                                    R|Plot method. Can be one of:
                                        {v}''', v='\n'.join(plot_func_names)))
    plot_parser.add_argument('capture_files', nargs='*',
                             help='One or more paths to capture files')
    for var_name, var_type in plot_func_param_data:
        plot_parser.add_argument(
            '--{}'.format(var_name.replace('_', '-')), type=var_type, required=False)

    args = parser.parse_args()

    if args.command == 'config':
        from tal.config import get_config_filename
        print('TAL configuration file is located at {l}'.format(
            l=get_config_filename()))
        print('Updating render backend configuration...')
        from tal.config import ask_for_config, Config
        version = ask_for_config(Config.MITSUBA_VERSION, force_ask=True)
        if version == '2':
            ask_for_config(
                Config.MITSUBA2_TRANSIENT_NLOS_FOLDER, force_ask=True)
        elif version == '3':
            ask_for_config(
                Config.MITSUBA3_TRANSIENT_NLOS_FOLDER, force_ask=True)
        else:
            raise AssertionError(
                f'Invalid MITSUBA_VERSION={version}, must be one of (2, 3)')
        print('Done.')
    elif args.command == 'render':
        config_file = args.config_file
        assert len(config_file) != 1 or config_file[0] != 'new', \
            'You must specify a folder name: tal render new <new_folder_name>'
        assert (len(config_file) == 1) or \
            (len(config_file) == 2 and config_file[0].lower() == 'new'), \
            'Usage: tal render new <new_folder_name> | tal render <config_file.yaml>"'
        if config_file[0].lower() == 'new':
            new_folder_name = config_file[1]
            from tal.render import create_nlos_scene
            create_nlos_scene(new_folder_name, args)
        else:
            from tal.render import render_nlos_scene
            config_file = config_file[0]
            render_nlos_scene(config_file, args)
    elif args.command == 'plot':
        import tal.plot
        from tal.io import read_capture
        command = args.preset
        assert command in plot_func_names, \
            'Unknown plot command: {}'.format(command)
        data = list()
        labels = []
        for capture_file in args.capture_files:
            print(f'Reading {capture_file}...')
            labels.append(capture_file)
            data.append(read_capture(capture_file))
        if len(data) == 1:
            data = data[0]
        if 'labels' in plot_func_param_names[command]:
            setattr(args, 'labels', labels)
        other_args = list(map(
            lambda p: None if p not in args else getattr(args, p),
            plot_func_param_names[command]))
        getattr(tal.plot, command)(data, *other_args)
    else:
        raise AssertionError('Invalid command? Check argparse')


if __name__ == '__main__':
    main()



from tal.enums import HFormat, GridFormat
from tal.config import get_resources
import numpy as np
from tqdm import tqdm


def backproject(H_0, laser_grid_xyz, sensor_grid_xyz, volume_xyz, volume_xyz_shape,
                camera_system, t_accounts_first_and_last_bounces,
                t_start, delta_t, is_confocal,
                projector_focus=None,
                laser_xyz=None, sensor_xyz=None, progress=False):

    if camera_system.is_transient():
        print('tal.reconstruct.bp: You have specified a time-resolved camera_system. '
              'The tal.reconstruct.bp implementation is better suited for time-gated systems. '
              'This will work, but you may want to check out tal.reconstruct.pf_dev for time-resolved reconstructions.')

    nt, nl, ns = H_0.shape
    nv, _ = volume_xyz.shape
    if is_confocal:
        assert laser_grid_xyz.shape[0] == ns, 'H does not match with laser_grid_xyz'
    else:
        assert laser_grid_xyz.shape[0] == nl, 'H does not match with laser_grid_xyz'
    assert sensor_grid_xyz.shape[0] == ns, 'H does not match with sensor_grid_xyz'
    assert (not t_accounts_first_and_last_bounces or (laser_xyz is not None and sensor_xyz is not None)), \
        't_accounts_first_and_last_bounces requires laser_xyz and sensor_xyz'

    # reshape everything into (nl, nv, ns, 3)
    if laser_xyz is not None:
        laser_xyz = laser_xyz.reshape((1, 1, 1, 3)).astype(np.float32)
    if sensor_xyz is not None:
        sensor_xyz = sensor_xyz.reshape((1, 1, 1, 3)).astype(np.float32)
    if is_confocal:
        laser_grid_xyz = laser_grid_xyz.reshape(
            (1, 1, ns, 3)).astype(np.float32)
    else:
        laser_grid_xyz = laser_grid_xyz.reshape(
            (nl, 1, 1, 3)).astype(np.float32)
    sensor_grid_xyz = sensor_grid_xyz.reshape((1, 1, ns, 3)).astype(np.float32)
    volume_xyz = volume_xyz.reshape((1, nv, 1, 3)).astype(np.float32)

    if camera_system.implements_projector():
        assert projector_focus is not None, 'projector_focus is required for this camera system'
        assert len(projector_focus) == 3, \
            'When using tal.reconstruct.bp, projector_focus must be a single 3D point. ' \
            'If you want to focus the illumination aperture at multiple points, ' \
            'please use tal.reconstruct.pf_dev instead or call tal.reconstruct.bp once per projector_focus.'
        if len(laser_grid_xyz) <= 3:
            print('tal.reconstruct.bp: You have specified a camera_system with a projector and a projector_focus, '
                  'but your data only contains one illumination point. Thus, you will not be able to implement the projector '
                  'i.e. focus the illumination aperture anywhere on the scene.')
        projector_focus = np.array(projector_focus).reshape(
            (1, 1, 1, 3)).repeat(nv, axis=1)
    else:
        assert projector_focus is None, \
            'projector_focus must not be set for this camera system'
        projector_focus = volume_xyz.reshape((1, nv, 1, 3))
    projector_focus = projector_focus.astype(np.float32)

    def distance(a, b):
        return np.linalg.norm(b - a, axis=-1)

    if camera_system.is_transient():
        H_1 = np.zeros((nt, *volume_xyz_shape), dtype=H_0.dtype)
    else:
        H_1 = np.zeros(volume_xyz_shape, dtype=H_0.dtype)

    # d_1: laser origin to laser illuminated point
    # d_2: laser illuminated point to projector_focus
    # d_3: x_v (camera_focus) to sensor imaged point
    # d_4: sensor imaged point to sensor origin
    if t_accounts_first_and_last_bounces:
        d_1 = distance(laser_xyz, laser_grid_xyz)
        d_4 = distance(sensor_grid_xyz, sensor_xyz)
    else:
        d_1 = np.float32(0.0)
        d_4 = np.float32(0.0)

    if camera_system.bp_accounts_for_d_2():
        d_2 = distance(laser_grid_xyz, projector_focus)
    else:
        d_2 = np.float32(0.0)

    def backproject_i(subrange_s):
        nsi = len(subrange_s)
        H_0_i = H_0[:, :, subrange_s]
        if is_confocal:
            d_2_i = d_2[:, :, subrange_s]
        else:
            d_2_i = d_2
        sensor_grid_xyz_i = sensor_grid_xyz[:, :, subrange_s, :]
        d_3 = distance(volume_xyz, sensor_grid_xyz_i)
        idx = d_1 + d_2_i + d_3 + d_4 - t_start
        idx /= delta_t
        idx = idx.astype(np.int32)

        t_range = nt if camera_system.is_transient() else 1
        t_range = np.arange(t_range, dtype=np.int32)
        if progress and len(t_range) > 1:
            t_range = tqdm(t_range,
                           desc='tal.reconstruct.bp time bins', leave=False)

        H_1_i = np.zeros((len(t_range), nv), dtype=H_0.dtype)
        i_v, i_s = np.ogrid[:nv, :nsi]
        for i_t in t_range:
            for i_l in range(nl):
                idx_i = idx[i_l, ...] + i_t
                good = np.logical_and(idx_i >= 0, idx_i < nt)
                idx_i[~good] = 0
                H_1_raw = H_0_i[idx_i[i_v, i_s], i_l, i_s]
                H_1_raw[~good] = 0.0
                H_1_i[i_t, :] += H_1_raw.sum(axis=1)

        if camera_system.is_transient():
            return H_1_i.reshape(volume_xyz_shape)
        else:
            return H_1_i[0].reshape(volume_xyz_shape)

    range_s = np.arange(ns, dtype=np.int32)
    h = H_0.dtype.itemsize
    s = sensor_grid_xyz.dtype.itemsize

    get_resources().split_work(
        backproject_i,
        data_in=range_s,
        data_out=H_1,
        slice_dims=(0, None),
    )

    return H_1



# pyright: reportMissingImports=false


def _get_setpath_location():
    from tal.config import ask_for_config, Config
    import os
    force_ask = False
    setpath_ok = False
    while not setpath_ok:
        setpath_location = os.path.join(
            ask_for_config(Config.MITSUBA3_TRANSIENT_NLOS_FOLDER,
                           force_ask=force_ask),
            'build', 'setpath.sh')
        if os.path.isfile(setpath_location):
            setpath_ok = True
        else:
            force_ask = True
            print(f'setpath.sh cannot be found in {setpath_location}.')
            print()
    return setpath_location


try:
    import mitsuba
except ModuleNotFoundError:
    import sys
    import subprocess
    setpath_location = _get_setpath_location()

    command = ['env', '-i', '/bin/bash',
               '-c', f'source {setpath_location} && printenv']
    p = subprocess.check_output(command).decode('utf-8')
    for line in p.split('\n')[:-1]:
        (key, _, value) = line.partition('=')
        if key == 'PYTHONPATH':
            for directory in value.split(':')[:-1]:
                sys.path.append(directory)


def get_name():
    return 'mitsuba3-transient-nlos'


def get_version():
    return '3.3.0'


def get_default_variant():
    return 'llvm_mono'


def set_variant(s):
    import mitsuba
    if not (s.startswith('llvm_') or s.startswith('cuda_')):
        raise AssertionError(
            f'Variant {s} is not supported. It must start with "llvm_" or "cuda_"')
    mitsuba.set_variant(s)


def get_hdr_extension():
    return 'npy'


def convert_hdr_to_ldr(hdr_path, ldr_path):
    from tal.util import write_img, tonemap_ldr
    image = _read_mitsuba_bitmap(hdr_path)
    image = tonemap_ldr(image)
    write_img(ldr_path, image)


def partial_laser_path(partial_results_dir, experiment_name, lx, ly):
    import os
    filename = os.path.join(
        partial_results_dir, f'{experiment_name}_L[{lx}][{ly}]'.replace('.', '_'))
    return f'{filename}.{get_hdr_extension()}', False


def read_transient_image(path):
    return _read_mitsuba_streakbitmap(path)


def remove_transient_image(path):
    import os
    if os.path.exists(path):
        os.remove(path)


def get_material_keys(s):
    import re
    return list(map(lambda e: e[1:],
                    re.compile(r'\$[a-zA-Z_]*').findall(s)))


def get_materials():
    from tal.util import fdent
    return {
        'white': fdent(f'''\
            <bsdf type="diffuse">
                <rgb name="reflectance" value="1.0, 1.0, 1.0"/>
            </bsdf>'''),
        'copper': fdent(f'''\
            <bsdf type="roughconductor">
                <string name="material" value="Cu"/>
                <string name="distribution" value="beckmann"/>
                <float name="alpha" value="$alpha"/>
            </bsdf>'''),
        'custom': '$text'
    }


def get_scene_xml(config, random_seed=0, quiet=False):
    import os
    import tal
    from tal.util import fdent

    def s(value):
        if isinstance(value, bool):
            return str(value).lower()
        else:
            return value

    def v(key):
        return s(config[key])

    # integrator
    integrator_steady = fdent(f'''\
        <integrator type="path"/>''')

    integrator_nlos = fdent(f'''\
        <integrator type="transient_nlos_path">
            <integer name="block_size" value="1"/>
            <integer name="max_depth" value="{v('integrator_max_depth')}"/>
            <integer name="filter_bounces" value="{v('integrator_filter_bounces')}"/>
            <boolean name="discard_direct_paths" value="{v('integrator_discard_direct_paths')}"/>
            <boolean name="nlos_laser_sampling" value="{v('integrator_nlos_laser_sampling')}"/>
            <boolean name="nlos_hidden_geometry_sampling" value="{v('integrator_nlos_hidden_geometry_sampling')}"/>
            <boolean name="nlos_hidden_geometry_sampling_do_rroulette" value="{v('integrator_nlos_hidden_geometry_sampling_do_rroulette')}"/>
            <boolean name="nlos_hidden_geometry_sampling_includes_relay_wall" value="{v('integrator_nlos_hidden_geometry_sampling_includes_relay_wall')}"/>
            <string name="temporal_filter" value="box"/>
        </integrator>''')

    if 'polarized' in v('mitsuba_variant'):
        # TODO(diego): mitsuba3 does not have a transient stokes integrator yet
        # https://github.com/mitsuba-renderer/mitsuba3/blob/master/src/integrators/stokes.cpp
        # https://github.com/diegoroyo/mitsuba2-transient-nlos/blob/a270850d1f9b9d863e759f880048df665cd7d2a1/src/integrators/transientstokes.cpp
        # https://github.com/diegoroyo/mitsuba3-transient-nlos/blob/main/mitransient/integrators/transientnlospath.py
        # the end goal would be to generate a transient stokes plugin in python (like transient_nlos_path.py)
        # that implements the functionality of stokes.cpp similar to mitsuba2-transient-nlos's transientstokes.cpp
        raise NotImplementedError(
            'Polarized variants are not implemented in mitsuba3 yet')

        def add_stokes(s, itype):
            return fdent('''\
                <integrator type="{itype}">
                    {s}
                </integrator>''', itype=itype, s=s)
        integrator_steady = add_stokes(integrator_steady, 'stokes')
        integrator_nlos = add_stokes(integrator_nlos, 'transientstokes')

    # relay wall
    geometry_names = list(map(lambda g: g['name'], v('geometry')))
    assert len(geometry_names) == len(set(geometry_names)), \
        'One or more geometry names is duplicated'
    relay_wall_name = v('relay_wall')
    assert relay_wall_name in geometry_names, \
        f'Relay wall is set to {relay_wall_name}, but must be one of {geometry_names}'

    dummy_lights_and_geometry_steady = fdent(f'''\
        <!-- Colored spheres to mark the laser and sensor positions -->
        <shape type="sphere">
            <point name="center" x="{v('sensor_x')}" y="{v('sensor_y')}" z="{v('sensor_z')}"/>
            <float name="radius" value="0.05"/>
            <bsdf type="diffuse" id="red">
                <rgb name="reflectance" value="1.0, 0.0, 0.0"/>
            </bsdf>
        </shape>
        <shape type="sphere">
            <point name="center" x="{v('laser_x')}" y="{v('laser_y')}" z="{v('laser_z')}"/>
            <float name="radius" value="0.05"/>
            <bsdf type="diffuse" id="blue">
                <rgb name="reflectance" value="0.0, 0.0, 1.0"/>
            </bsdf>
        </shape>

        <!-- Illuminate all the scene -->
        <emitter type="point">
            <rgb name="intensity" value="1.0, 1.0, 1.0"/>
            <point name="position" x="30" y="0" z="30"/>
        </emitter>
        <emitter type="point">
            <rgb name="intensity" value="1.0, 1.0, 1.0"/>
            <point name="position" x="-30" y="0" z="30"/>
        </emitter>
        <emitter type="point">
            <rgb name="intensity" value="1.0, 1.0, 1.0"/>
            <point name="position" x="0" y="30" z="30"/>
        </emitter>
        <emitter type="point">
            <rgb name="intensity" value="1.0, 1.0, 1.0"/>
            <point name="position" x="0" y="-30" z="30"/>
        </emitter>''')

    pixel_format = 'luminance' if 'mono' in v('mitsuba_variant') else 'rgb'

    sensors_steady = fdent(f'''\
        <!-- Sensor 0: back view -->
        <sensor type="perspective">
            <string name="fov_axis" value="smaller"/>
            <float name="near_clip" value="0.01"/>
            <float name="far_clip" value="1000"/>
            <float name="fov" value="30"/>
            <transform name="to_world">
                <lookat origin="0, 0, 5"
                        target="0, 0, 0"
                            up="0, 1, 0"/>
            </transform>
            <sampler type="independent">
                <integer name="sample_count" value="512"/>
                <integer name="seed" value="{random_seed}"/>
            </sampler>
            <film type="hdrfilm">
                <integer name="width" value="512"/>
                <integer name="height" value="512"/>
                <string name="pixel_format" value="{pixel_format}"/>
                <rfilter name="rfilter" type="gaussian"/>
            </film>
        </sensor>

        <!-- Sensor 1: side view -->
        <sensor type="perspective">
            <string name="fov_axis" value="smaller"/>
            <float name="near_clip" value="0.01"/>
            <float name="far_clip" value="1000"/>
            <float name="fov" value="45"/>
            <transform name="to_world">
                <lookat origin="5, 0, 1.5"
                        target="0, 0, 1.5"
                            up="0, 1, 0"/>
            </transform>
            <sampler type="independent">
                <integer name="sample_count" value="512"/>
            </sampler>
            <film type="hdrfilm">
                <integer name="width" value="512"/>
                <integer name="height" value="512"/>
                <string name="pixel_format" value="{pixel_format}"/>
                <rfilter name="rfilter" type="gaussian"/>
            </film>
        </sensor>''')

    confocal_capture = 'false'
    if v('scan_type') == 'confocal':
        confocal_capture = 'true'
    elif v('scan_type') != 'single' and v('scan_type') != 'exhaustive':
        raise AssertionError(
            'scan_type should be one of {single|confocal|exhaustive}')
    sensor_nlos = fdent(f'''\
        <sensor type="nlos_capture_meter">
            <sampler type="independent">
                <integer name="sample_count" value="{v('sample_count')}"/>
                <integer name="seed" value="{random_seed}"/>
            </sampler>

            <emitter type="projector">
                <rgb name="irradiance" value="1.0, 1.0, 1.0"/>
                <float name="fov" value="{0.2 if v('integrator_nlos_laser_sampling') else 2}"/>
            </emitter>

            <boolean name="confocal" value="{confocal_capture}"/>
            <boolean name="account_first_and_last_bounces" value="{v('account_first_and_last_bounces')}"/>
            <point name="sensor_origin" x="{v('sensor_x')}" y="{v('sensor_y')}" z="{v('sensor_z')}"/>
            <point name="laser_origin" x="{v('laser_x')}" y="{v('laser_y')}" z="{v('laser_z')}"/>
            <point name="laser_lookat_pixel" x="$laser_lookat_x" y="$laser_lookat_y" z="0"/>
            <film type="transient_hdr_film">
                <integer name="width" value="{v('sensor_width')}"/>
                <integer name="height" value="{v('sensor_height')}"/>

                <integer name="temporal_bins" value="{v('num_bins')}"/>
                <!-- <boolean name="auto_detect_bins" value="{v('auto_detect_bins')}"/> -->
                <float name="bin_width_opl" value="{v('bin_width_opl')}"/>
                <float name="start_opl" value="{v('start_opl')}"/>
                <rfilter type="box">
                    <!-- <float name="radius" value="0.5"/> -->
                </rfilter>
            </film>
        </sensor>''')

    materials = get_materials()
    shapes_steady = []
    shapes_nlos = []
    for gdict in v('geometry'):
        def g(key):
            return s(gdict.get(key, None))

        name = g('name')
        is_relay_wall = name == relay_wall_name
        if is_relay_wall and g('mesh')['type'] != 'rectangle' and not quiet:
            print('WARNING: Relay wall does not work well with meshes that are '
                  'not of type "rectangle" because of wrong UV mapping. '
                  'Please make sure that you know what you are doing')
        shape_name = f'<!-- {name}{" (RELAY WALL)" if is_relay_wall else ""} -->'
        description = g('description')
        if description is not None and len(description.strip()) > 0:
            shape_name = fdent('''\
                {shape_name}
                <!--
                    {description}
                -->''', shape_name=shape_name, description=description.strip())

        material_id = g('material')['id']
        assert material_id in materials.keys(), \
            f'Geometry {name} has material {material_id}, but it must be one of {", ".join(materials.keys())}'
        shape_material = materials[material_id]
        material_keys = get_material_keys(shape_material)
        for key in material_keys:
            assert key in g('material'), \
                f'Material {material_id} for geometry {name} must have a value for {key}'
            shape_material = shape_material.replace(
                f'${key}', g('material')[key])

        shape_transform = fdent(f'''\
            <transform name="to_world">
                <scale x="{g('scale') or 1.0}" y="{g('scale') or 1.0}" z="{g('scale') or 1.0}"/>
                <rotate x="1" angle="{g('rot_degrees_x') or 0.0}"/>
                <rotate y="1" angle="{g('rot_degrees_y') or 0.0}"/>
                <rotate z="1" angle="{g('rot_degrees_z') or 0.0}"/>
                <translate x="{g('displacement_x') or 0.0}" y="{g('displacement_y') or 0.0}" z="{g('displacement_z') or 0.0}"/>
            </transform>''')

        shape_contents_steady = f'{shape_material}\n{shape_transform}'
        newline = '\n'
        shape_contents_nlos = f'{shape_material}\n{shape_transform}{f"{newline}{sensor_nlos}" if is_relay_wall else ""}'

        if g('mesh')['type'] == 'obj':
            assert 'filename' in g('mesh'), \
                f'Missing mesh filename for geometry "{name}". ' \
                f'It is required because its mesh type is set as OBJ.'
            filename = g('mesh')['filename']
            assert os.path.isfile(filename), \
                f'{filename} does not exist for geometry "{name}"'

            def shapify(content, filename):
                return fdent('''\
                {shape_name}
                <shape type="obj">
                    <string name="filename" value="{filename}"/>

                    {content}
                </shape>''', shape_name=shape_name, filename=filename, content=content)

            shapes_steady.append(shapify(shape_contents_steady, filename))
            shapes_nlos.append(shapify(shape_contents_nlos, filename))
        elif g('mesh')['type'] == 'rectangle' or g('mesh')['type'] == 'sphere':
            def shapify(content):
                return fdent('''\
                {shape_name}
                <shape type="{shape_type}">
                    {content}
                </shape>''', shape_name=shape_name, content=content,
                             shape_type=g('mesh')['type'])

            shapes_steady.append(shapify(shape_contents_steady))
            shapes_nlos.append(shapify(shape_contents_nlos))
        else:
            raise AssertionError(
                f'Shape not yet supported: {g("mesh")["type"]}')

    shapes_steady = '\n\n'.join(shapes_steady)
    shapes_nlos = '\n\n'.join(shapes_nlos)

    file_steady = fdent('''\
        <!-- Auto-generated using TAL v{tal_version} -->
        <scene version="{mitsuba_version}">
            {integrator_steady}

            {dummy_lights_and_geometry_steady}

            {shapes_steady}

            {sensors_steady}
        </scene>''',
                        tal_version=tal.__version__,
                        mitsuba_version=get_version(),
                        integrator_steady=integrator_steady,
                        dummy_lights_and_geometry_steady=dummy_lights_and_geometry_steady,
                        shapes_steady=shapes_steady,
                        sensors_steady=sensors_steady)

    file_nlos = fdent('''\
        <!-- Auto-generated using TAL v{tal_version} -->
        <scene version="{mitsuba_version}">
            {integrator_nlos}

            {shapes_nlos}
        </scene>''',
                      tal_version=tal.__version__,
                      mitsuba_version=get_version(),
                      integrator_nlos=integrator_nlos,
                      shapes_nlos=shapes_nlos)

    return file_steady, file_nlos


def run_mitsuba(scene_xml_path, hdr_path, defines,
                experiment_name, logfile, args, sensor_index=0, queue=None):
    try:
        import mitsuba as mi
        import mitransient as mitr
        import numpy as np
        from tqdm import tqdm
        from mitransient.integrators.common import TransientADIntegrator
        import sys
        import os

        sys.stdout = queue
        sys.stderr = queue
        os.nice(args.nice)

        if args.dry_run:
            return

        scene = mi.load_file(scene_xml_path, **defines)
        integrator = scene.integrator()

        mi.Thread.set_thread_count(args.threads)

        # prepare
        if isinstance(integrator, TransientADIntegrator):
            integrator.prepare_transient(scene, sensor_index)

            progress_bar = None
            if not args.quiet:
                progress_bar = tqdm(total=100, desc=experiment_name,
                                    ascii=True, leave=False)

            def update_progress(p):
                if not args.quiet:
                    progress_bar.n = int(p * 100)
                    progress_bar.refresh()

            steady_image, transient_image = integrator.render(
                scene, progress_callback=update_progress)
            result = np.array(transient_image)
            if result.ndim == 2:
                nt, nc = result.shape
                result = result.reshape((nt, 1, 1, nc))
            result = np.moveaxis(result, 2, 0)
            # result has shape (nt, nx, ny, nchannels)
            if result.ndim == 4:
                # sum all channels
                result = np.sum(result, axis=-1)
            if not args.quiet:
                progress_bar.close()
            del steady_image, transient_image, progress_bar
        else:
            image = integrator.render(scene, sensor_index)
            result = np.array(image)

        np.save(hdr_path, result)
        del result, scene, integrator
    except Exception as e:
        queue.write(e)


def _read_mitsuba_bitmap(path: str):
    import numpy as np
    return np.load(path)


def _read_mitsuba_streakbitmap(path: str, exr_format='RGB'):
    import numpy as np
    return np.load(path)



from tal.enums import HFormat, GridFormat, VolumeFormat, CameraSystem
from tal.io.capture_data import NLOSCaptureData
from typing import Union
import numpy as np


def _infer_volume_format(volume_xyz, volume_format, log=True):
    # infer volume_format if it is unknown
    if volume_format == VolumeFormat.UNKNOWN:
        assert volume_xyz.shape[-1] == 3, \
            'Could not infer volume_format. Please specify it manually.'
        if volume_xyz.ndim == 2:
            if log:
                print('tal.reconstruct.utils: Assuming that volume_xyz is N_3')
            volume_format = VolumeFormat.N_3
        elif volume_xyz.ndim == 3:
            if log:
                print('tal.reconstruct.utils: Assuming that volume_xyz is X_Y_3')
            volume_format = VolumeFormat.X_Y_3
        elif volume_xyz.ndim == 4:
            if log:
                print('tal.reconstruct.utils: Assuming that volume_xyz is X_Y_Z_3')
            volume_format = VolumeFormat.X_Y_Z_3
        else:
            raise AssertionError(
                'Could not infer volume_format. Please specify it manually.')

    return volume_format


def convert_to_N_3(data: NLOSCaptureData,
                   volume_xyz: NLOSCaptureData.VolumeXYZType,
                   volume_format: VolumeFormat = VolumeFormat.UNKNOWN,
                   try_optimize_convolutions: bool = False):
    """
        try_optimize_convolutions performs many checks:
        - H, {laser|sensor}_grid_xyz and volume_xyz have X, Y components (e.g. they are not N_3)
        - The {laser|sensor}_grid_xyz's slices are parallel to the volume_xyz's slices
        - The points in the {laser|sensor}_grid_xyz's slices are sampled at the same rate
    """

    volume_format = _infer_volume_format(volume_xyz, volume_format, log=True)

    # this variable is set to false during the conversion
    optimize_projector_convolutions = try_optimize_convolutions and \
        volume_format in [VolumeFormat.X_Y_3, VolumeFormat.X_Y_Z_3]
    optimize_camera_convolutions = optimize_projector_convolutions

    is_confocal = data.is_confocal()

    if data.H_format == HFormat.T_Si:
        optimize_projector_convolutions = False
        optimize_camera_convolutions = False
        nt, ns = data.H.shape
        if is_confocal:
            nl = ns
        H = data.H.reshape(nt, 1, ns)
    elif data.H_format == HFormat.T_Sx_Sy:
        optimize_projector_convolutions = False
        nt, nsx, nsy = data.H.shape
        if is_confocal:
            nlx, nly = nsx, nsy
        H = data.H.reshape(nt, 1, nsx * nsy)
    elif data.H_format == HFormat.T_Li_Si:
        optimize_projector_convolutions = False
        optimize_camera_convolutions = False
        nt, nl, ns = data.H.shape
        H = data.H.reshape(nt, nl, ns)
    elif data.H_format == HFormat.T_Lx_Ly_Sx_Sy:
        nt, nlx, nly, nsx, nsy = data.H.shape
        H = data.H.reshape(nt, nlx * nly, nsx * nsy)
    else:
        raise AssertionError(f'H_format {data.H_format} not implemented')

    if data.laser_grid_format == GridFormat.N_3:
        optimize_projector_convolutions = False
        laser_grid_xyz = data.laser_grid_xyz
    elif data.laser_grid_format == GridFormat.X_Y_3:
        try:
            assert nlx == data.laser_grid_xyz.shape[0] and nly == data.laser_grid_xyz.shape[1], \
                'laser_grid_xyz.shape does not match with H.shape'
        except NameError:
            # nlx, nly not defined, OK
            nlx, nly, _ = data.laser_grid_xyz.shape
            pass
        laser_grid_xyz = data.laser_grid_xyz.reshape(nlx * nly, 3)
        optimize_projector_convolutions = optimize_projector_convolutions and nlx > 1 and nly > 1
    else:
        raise AssertionError(
            f'laser_grid_format {data.laser_grid_format} not implemented')

    if data.sensor_grid_format == GridFormat.N_3:
        optimize_camera_convolutions = False
        sensor_grid_xyz = data.sensor_grid_xyz
    elif data.sensor_grid_format == GridFormat.X_Y_3:
        try:
            assert nsx == data.sensor_grid_xyz.shape[0] and nsy == data.sensor_grid_xyz.shape[1], \
                'sensor_grid_xyz.shape does not match with H.shape'
        except NameError:
            # nsx, nsy not defined, OK
            nsx, nsy, _ = data.sensor_grid_xyz.shape
            pass
        sensor_grid_xyz = data.sensor_grid_xyz.reshape(nsx * nsy, 3)
        optimize_camera_convolutions = optimize_camera_convolutions and nsx > 1 and nsy > 1
    else:
        raise AssertionError(
            f'sensor_grid_format {data.sensor_grid_format} not implemented')

    if is_confocal:
        assert H.shape[1] == 1 and H.shape[2] == laser_grid_xyz.shape[0], \
            'H.shape does not match with laser_grid_xyz.shape.'
    else:
        assert H.shape[1] == laser_grid_xyz.shape[0], \
            'H.shape does not match with laser_grid_xyz.shape.'
    assert H.shape[2] == sensor_grid_xyz.shape[0], \
        'H.shape does not match with sensor_grid_xyz.shape.'

    assert volume_format.xyz_dim_is_last(), 'Unexpected volume_format'

    if volume_format in [VolumeFormat.X_Y_3, VolumeFormat.X_Y_Z_3]:
        # list of (Z, 3) normals of all Z positions in the plane
        if volume_format == VolumeFormat.X_Y_3:
            z_index = Ellipsis
        else:
            z_index = 0
        v_a = volume_xyz[0, 0, z_index, :]
        v_b = volume_xyz[-1, 0, z_index, :]
        v_c = volume_xyz[0, -1, z_index, :]
        v_n = np.cross(v_b - v_a, v_c - v_a).reshape((-1, 3))
        v_n /= np.linalg.norm(v_n, axis=-1, keepdims=True)

        v_dx = np.linalg.norm(
            volume_xyz[1, 0, z_index, :] - volume_xyz[0, 0, z_index, :])
        v_dy = np.linalg.norm(
            volume_xyz[0, 1, z_index, :] - volume_xyz[0, 0, z_index, :])
    else:
        optimize_projector_convolutions = False
        optimize_camera_convolutions = False

    try:
        assert optimize_projector_convolutions
        assert data.laser_grid_format == GridFormat.X_Y_3
        laser_grid_xyz = laser_grid_xyz.reshape((nlx, nly, 3))
        l_a = laser_grid_xyz[0, 0, ..., :]
        l_b = laser_grid_xyz[-1, 0, ..., :]
        l_c = laser_grid_xyz[0, -1, ..., :]
        l_n = np.cross(l_b - l_a, l_c - l_a).reshape((1, 3))
        l_n /= np.linalg.norm(l_n, axis=-1, keepdims=True)

        l_dx = np.linalg.norm(
            laser_grid_xyz[1, 0, ..., :] - laser_grid_xyz[0, 0, ..., :])
        l_dy = np.linalg.norm(
            laser_grid_xyz[0, 1, ..., :] - laser_grid_xyz[0, 0, ..., :])

        dot_lv = np.sum(l_n * v_n, axis=-1)
        assert np.allclose(np.abs(dot_lv), 1)
        assert np.isclose(v_dx, l_dx) and np.isclose(v_dy, l_dy)
        print('tal.reconstruct.utils: Optimizing for projector convolutions.')
    except AssertionError:
        optimize_projector_convolutions = False
        laser_grid_xyz = laser_grid_xyz.reshape((-1, 3))

    try:
        assert optimize_camera_convolutions
        assert data.sensor_grid_format == GridFormat.X_Y_3
        sensor_grid_xyz = sensor_grid_xyz.reshape((nsx, nsy, 3))
        s_a = sensor_grid_xyz[0, 0, ..., :]
        s_b = sensor_grid_xyz[-1, 0, ..., :]
        s_c = sensor_grid_xyz[0, -1, ..., :]
        s_n = np.cross(s_b - s_a, s_c - s_a).reshape((1, 3))
        s_n /= np.linalg.norm(s_n, axis=-1, keepdims=True)

        s_dx = np.linalg.norm(
            sensor_grid_xyz[1, 0, ..., :] - sensor_grid_xyz[0, 0, ..., :])
        s_dy = np.linalg.norm(
            sensor_grid_xyz[0, 1, ..., :] - sensor_grid_xyz[0, 0, ..., :])

        dot_sv = np.sum(s_n * v_n, axis=-1)
        assert np.allclose(np.abs(dot_sv), 1)
        assert np.isclose(v_dx, s_dx) and np.isclose(v_dy, s_dy)
        print('tal.reconstruct.utils: Optimizing for camera convolutions.')
    except AssertionError:
        optimize_camera_convolutions = False
        sensor_grid_xyz = sensor_grid_xyz.reshape((-1, 3))

    if not optimize_projector_convolutions and not optimize_camera_convolutions:
        volume_xyz = volume_xyz.reshape((-1, 3))

    return (H, laser_grid_xyz, sensor_grid_xyz, volume_xyz,
            optimize_projector_convolutions, optimize_camera_convolutions)


def convert_reconstruction_from_N_3(data: NLOSCaptureData,
                                    reconstructed_volume_n3: Union[NLOSCaptureData.SingleReconstructionType,
                                                                   NLOSCaptureData.ExhaustiveReconstructionType],
                                    volume_xyz: NLOSCaptureData.VolumeXYZType,
                                    volume_format: VolumeFormat,
                                    camera_system: CameraSystem,
                                    is_exhaustive_reconstruction: bool = False):

    volume_format = _infer_volume_format(volume_xyz, volume_format, log=False)

    if camera_system.is_transient():
        shape = (data.H.shape[data.H_format.time_dim()],)
    else:
        shape = ()

    assert volume_format.xyz_dim_is_last(), 'Unexpected volume_format'

    if is_exhaustive_reconstruction:
        # add an additional shape dimension for the exhaustive reconstruction
        shape += volume_xyz.shape[:-1]
    shape += volume_xyz.shape[:-1]

    return reconstructed_volume_n3.reshape(shape)



from enum import Enum
import os
import h5py
import yaml
import numpy as np
import inspect
from typing import Union, get_type_hints
from nptyping import NDArray, Shape
from tal.io.format import convert_dict, detect_dict_format
from tal.enums import FileFormat, HFormat, GridFormat, VolumeFormat


class LazyDataset:
    def __init__(self, dataset):
        self.read = False
        self.dataset = dataset

    def __get__(self, instance, owner):
        if not self.read:
            self.dataset = self.dataset[()]
            self.read = True
        return self.dataset

    def __getattr__(self, name):
        return getattr(self.dataset, name)


def read_hdf5(filename: str) -> dict:
    raw_data = h5py.File(filename, 'r')

    def parse(key, value):
        if isinstance(value, h5py.Empty):
            value = None
        elif isinstance(value, h5py.Group):
            value = {k: parse(k, v) for k, v in value.items()}
        else:
            if isinstance(value, h5py.Dataset):
                value = value[()]
                # FIXME(diego): unused, need to figure out how
                # to call do stuff like value.size or
                # isinstance(value, bytes) without reading the
                # whole dataset
                # value = LazyDataset(value)
            if isinstance(value, np.ndarray) and value.size == 1:
                value = np.squeeze(value)[()]

        # parse enums and strings
        if key in get_type_hints(NLOSCaptureData):
            key_class = get_type_hints(NLOSCaptureData)[key]
            if inspect.isclass(key_class) and issubclass(key_class, Enum):
                if isinstance(value, h5py.Empty):
                    value = key_class(0)
                else:
                    value = key_class(value)
        elif isinstance(value, bytes):
            # FIXME(diego): not working?
            value = yaml.safe_load(value)
        return value

    raw_data = dict((key, parse(key, raw_data.get(key)))
                    for key in raw_data.keys())
    return raw_data


def write_hdf5(capture_data: dict, filename: str):
    file = h5py.File(filename, 'w')
    for key, value in capture_data.items():
        if value is None:
            file[key] = h5py.Empty(float)
        elif isinstance(value, dict):
            file[key] = yaml.dump(value)
        elif isinstance(value, Enum):
            dt = h5py.enum_dtype(dict((item.name, item.value)
                                 for item in value.__class__), basetype='i')
            ds = file.create_dataset(key, (1,), dtype=dt)
            ds[0] = value.value
        else:
            file[key] = value
    file.close()


class NLOSCaptureData:
    """
    Contains data for a NLOS capture scene. Typically, a laser illuminates a planar wall with a delta light pulse.
    An ultrafast camera captures the reflected light at that wall.

    The scene is described by the following parameters:

    H
        Multi-dimensional (>=2 dimensions) impulse response function.
        Contains time-resolved illumination for different laser and sensor positions.
        See H_format for the number of dimensions and what each one contains.

    H_format
        Enum describing what each dimension contains.
        e.g. HFormat.T_Lx_Ly_Sx_Sy means that there is temporal (T) data for a 2D grid
             of laser positions (Lx, Ly) and a 2D grid of sensor positions (Sx, Sy)
        e.g. HFormat.T_Sx_Sy means that there is temporal (T) data for a 2D grid of sensor positions.
             In this case, see is_confocal() to differentiate between the case of only one laser point
             and the case of a 2D grid of laser positions co-located with the sensor positions.
             You can also check sensor_grid_xyz and laser_grid_xyz, but is_confocal() is easier.

    sensor_xyz, laser_xyz
        3D coordinates of the sensor and laser positions.
        Note that it is the origin of the laser itself, and _not_ the point where the laser is focused at.

    sensor_grid_xyz, laser_grid_xyz
        Positions of laser and sensor points at the wall.
        See {sensor|laser}_grid_format for the number of dimensions.

    sensor_grid_normals, laser_grid_normals
        Normal vectors of the laser and sensor points at the wall.
        See {sensor|laser}_grid_format for the number of dimensions.

    sensor_grid_format, laser_grid_format
        Enum describing what each dimension of {sensor|laser}_grid_{xyz|normals} contains.
        See tal.enums.GridFormat for details.

    delta_t
        Temporal resolution of H i.e. distance between two time bins.
        Measured in meters (see https://en.wikipedia.org/wiki/Optical_path_length).

    t_start
        Initial instant of H, used to define a different time origin.
        Measured in meters (see https://en.wikipedia.org/wiki/Optical_path_length).

    t_accounts_first_and_last_bounces
        Whether the impulse response accounts for the time-of-flight between
        the laser origin and the wall (first bounce),
        and the wall and the sensor origin (last bounce).

    scene_info
        YAML-encoded string. Contains additional information about the scene. Implemented keys:
        - 'original_format': str (e.g. 'HDF5_ZNLOS')
        - 'config': dict (original scene_config.yaml when generated using TAL)
        - 'args': dict (original args passed when generated using TAL)
        - 'volume': dict
            - 'center': Array3 (center of volume)
            - 'rotation': Array3 (rotation of volume formatted as per Z-NLOS - probably unused for now)
            - 'size': Array3 (size of volume - length of each XYZ side)
            - 'xyz': MatrixN3 (if ever volume_format is VolumeFormat.N_3, points will be stored here)
    """

    #
    # Type aliases
    #
    Float = np.float32
    TensorTSxSy = NDArray[Shape['T, Sx, Sy'], Float]
    TensorTLxLySxSy = NDArray[Shape['T, Lx, Ly, Sx, Sy'], Float]
    HType = Union[TensorTSxSy, TensorTLxLySxSy]
    MatrixN3 = NDArray[Shape['*, 3'], Float]
    TensorXY3 = NDArray[Shape['X, Y, 3'], Float]
    LaserGridType = Union[MatrixN3, TensorXY3]
    SensorGridType = Union[MatrixN3, TensorXY3]
    TensorXYZ3 = NDArray[Shape['X, Y, Z, 3'], Float]
    VolumeXYZType = Union[MatrixN3, TensorXY3, TensorXYZ3]
    Array3 = NDArray[Shape['3'], Float]

    # reconstruction types
    TensorN = NDArray[Shape['N'], Float]
    TensorXY = NDArray[Shape['X, Y'], Float]
    TensorXYZ = NDArray[Shape['X, Y, Z'], Float]
    SingleReconstructionType = Union[TensorN, TensorXY, TensorXYZ]
    TensorNN = NDArray[Shape['N, N'], Float]
    TensorXYXY = NDArray[Shape['X, Y, X, Y'], Float]
    TensorXYZXYZ = NDArray[Shape['X, Y, Z, X, Y, Z'], Float]
    ExhaustiveReconstructionType = Union[TensorNN, TensorXYXY, TensorXYZXYZ]

    #
    # Actual capture data (ignore _start and _end)
    #
    _start: None = None  # used in as_dict()
    H: HType = None
    H_format: HFormat = None
    sensor_xyz: Array3 = None
    sensor_grid_xyz: SensorGridType = None
    sensor_grid_normals: SensorGridType = None
    sensor_grid_format: GridFormat = None
    laser_xyz: Array3 = None
    laser_grid_xyz: LaserGridType = None
    laser_grid_normals: LaserGridType = None
    laser_grid_format: GridFormat = None
    # volume_xyz : VolumeXYZType = None  # deprecated
    volume_format: VolumeFormat = None  # deprecated
    delta_t: Float = None
    t_start: Float = None
    t_accounts_first_and_last_bounces: bool = None
    scene_info: dict = None  # additional information
    _end: None = None  # used in as_dict()

    def __get_dict_keys(self):
        variables = list(self.__class__.__dict__.keys())
        variables = variables[variables.index(
            '_start') + 1: variables.index('_end')]
        return variables

    def __init__(self, filename: str = None, file_format: FileFormat = FileFormat.AUTODETECT):
        if filename is None:
            return

        # Read raw data and check its current format
        assert os.path.isfile(filename), f'Does not exist: {filename}'
        raw_data = read_hdf5(filename)
        if file_format == FileFormat.AUTODETECT:
            file_format = detect_dict_format(raw_data)

        # Convert data to HDF5_TAL format
        if file_format != FileFormat.HDF5_TAL:
            raw_data = convert_dict(
                raw_data, format_to=FileFormat.HDF5_TAL)

        own_dict_keys = self.__get_dict_keys()
        for key, value in raw_data.items():
            if key not in own_dict_keys:
                raise AssertionError(f'raw_data contains unknown key: {key}')
            setattr(self, key, value)

    def is_confocal(self):
        """ Returns True if H contains confocal data (i.e. Sx and Sy represent both laser and sensor coordinates) """
        if self.H_format in [HFormat.T_Lx_Ly_Sx_Sy, HFormat.T_Li_Si]:
            return False
        elif self.H_format in [HFormat.T_Sx_Sy, HFormat.T_Si]:
            return np.allclose(self.sensor_grid_xyz, self.laser_grid_xyz)
        else:
            raise AssertionError('Invalid H_format')

    def as_dict(self):
        """ Returns a dict containing all the data in this object """
        dict_keys = self.__get_dict_keys()
        return dict((key, getattr(self, key)) for key in dict_keys)

    def downscale(self, downscale: int):
        """
        Updates the data in this object to reduce the number of laser and sensor positions by the given factor

        e.g. Consider data with H_format = (T, Sx, Sy) and H.shape = (4096, 256, 256).
             Calling with downscale = 2 will reduce H.shape to (4096, 128, 128).
        """
        assert downscale > 1, 'downscale must be > 1'
        assert self.H_format in [HFormat.T_Sx_Sy], \
            'Only implemented for HFormat.T_Sx_Sy'
        nt, nsx, nsy = self.H.shape
        self.H = self.H.reshape(
            (nt, nsx // downscale, downscale, nsy // downscale, downscale)).sum(axis=(2, 4))
        self.sensor_grid_xyz = self.sensor_grid_xyz.reshape(
            (nsx // downscale, downscale, nsy // downscale, downscale, 3)).mean(axis=(1, 3))
        print(
            f'Downscaled from {nsx}x{nsy} to {nsx // downscale}x{nsy // downscale}')
